# -*- coding: utf-8 -*-

import functools
import itertools

from django import forms

from .lomake import yhdista_lomakkeet
from .sarja import lisaa_lomakesarja


def nido(*args, **kwargs):
  if len(args) > 2:
    return functools.reduce(functools.partial(nido, **kwargs), args)

  elif len(args) == 1:
    def koriste(lomake_a):
      return nido(lomake_a, *args, **kwargs)
    return koriste

  lomake_a, lomake_b = args
  assert isinstance(lomake_a, type) and issubclass(lomake_a, forms.ModelForm), \
  f'Virheellinen A-lomakeluokka {lomake_a!r}'
  assert isinstance(lomake_b, type) and issubclass(lomake_b, forms.ModelForm), \
  f'Virheellinen B-lomakeluokka {lomake_b!r}'

  if 'avain_a' in kwargs:
    avain_a = kwargs.pop('avain_a')
  else:
    avain_a, *_ = itertools.chain((
      f.name for f in lomake_a.Meta.model._meta.get_fields()
      if f.is_relation and f.related_model == lomake_b.Meta.model
    ), (None, ))
  if avain_a is not None:
    kwargs['avain_a'] = avain_a

  if 'avain_b' in kwargs:
    avain_b = kwargs.pop('avain_b')
  else:
    avain_b, *_ = itertools.chain((
      f.name for f in lomake_b.Meta.model._meta.get_fields()
      if f.is_relation and f.related_model == lomake_a.Meta.model
    ), (None, ))
  if avain_b is not None:
    kwargs['avain_b'] = avain_b

  useita = kwargs.pop('useita', None)
  if useita is None:
    if avain_a is not None:
      viittaus = lomake_a.Meta.model._meta.get_field(avain_a)
      useita = viittaus.one_to_many or viittaus.many_to_many
    elif avain_b is not None:
      viittaus = lomake_b.Meta.model._meta.get_field(avain_b)
      useita = viittaus.many_to_one or viittaus.many_to_many
    else:
      raise ValueError(
        'Mallien välisen viittaussuhteen lukuisuutta ei voitu päätellä.'
        ' Joko `avain_a` tai `avain_b` on oltava `!= None`.'
      )
    # if useita is None

  if useita:
    from django.contrib.contenttypes.fields import GenericForeignKey
    kwargs.setdefault('epasuora', isinstance(
      lomake_b.Meta.model._meta.get_field(avain_b),
      GenericForeignKey,
    ) if avain_b is not None else False)

  return (lisaa_lomakesarja if useita else yhdista_lomakkeet)(*args, **kwargs)
  # def nido



# -*- coding: utf-8 -*-
# pylint: disable=invalid-name

from collections.abc import Iterable
import functools
from itertools import chain
import types

from django.db import DatabaseError, transaction
from django import forms
from django.utils.functional import cached_property
from django.utils.module_loading import import_string

from .nidottu import yhdistetty_lomake
from .piirto import Piirto


@functools.wraps(yhdistetty_lomake)
def yhdista_lomakkeet(
  LomakeA: type,
  LomakeB: type,
  *,
  avain_a: str = None,
  avain_b: str = None,
  tunnus: str = None,
  pakollinen_b: bool = None,
  valita_parametrit: Iterable = (),
):
  '''
  Yhdistää kaksi ModelForm-luokkaa silloin,
  kun A-luokasta on suora viittaus B-luokkaan

  Joko `avain_a` tai `avain_b` on pakollinen.

  Args:
    LomakeA, LomakeB: mallikohtaiset lomakeluokat
    avain_a: viittausavain LomakeA.Meta.model-luokan kohteelta
      LomakeB.Meta.model-luokan kohteelle
    avain_b: viittausavain LomakeB.Meta.model-luokan kohteelta
      LomakeA.Meta.model-luokan kohteelle
      (oletuksena haetaan viittauksen vastaluokan tiedoista)
    tunnus: yksikäsitteinen tunnus LomakeB-lomakkeelle
      (oletuksena ``avain_a`` tai ``avain_b``:n vastaviittaus)
    pakollinen_b: onko lomake B pakollinen, tarkistettava tieto?
      (oletuksena sama kuin A--B-viittauksen pakollisuus, jos annettu)
  '''

  if isinstance(LomakeA, str):
    LomakeA = import_string(LomakeA)
  if isinstance(LomakeB, str):
    LomakeB = import_string(LomakeB)

  # Muodosta avain_a, avain_b, kentta_a ja kentta_b.
  assert avain_a or avain_b
  kentta_a = LomakeA.Meta.model._meta.get_field(avain_a) if avain_a else None

  avain_b = avain_b or kentta_a.remote_field.name
  kentta_b = LomakeB.Meta.model._meta.get_field(avain_b)
  tunnus = tunnus or avain_a or kentta_b.remote_field.name

  if tunnus in LomakeA.base_fields:
    raise RuntimeError(
      f'Sisempää lomaketta ei voida nitoa samalla nimellä'
      f' kuin olemassaoleva lomakekenttä: {tunnus}'
    )

  # Mikäli A-avainta ei ole annettu,
  # B-lomake on oletuksena valinnainen.
  # Muutoin haetaan oletus vastaavan kentän tiedoista.
  if pakollinen_b is None:
    pakollinen_b = avain_a and (
      not LomakeA.Meta.model._meta.get_field(avain_a).null
    )

  def liitos_kwargs(self: LomakeA):
    kohde_b = self.initial.get(tunnus, None)
    if isinstance(kohde_b, LomakeB.Meta.model):
      pass
    elif kohde_b is not None:
      kohde_b = LomakeB.Meta.model.objects.get(pk=kohde_b)
    elif self.instance and avain_a:
      kohde_b = getattr(self.instance, avain_a, None)

    # Mikäli olemassaolevaa B-kohdetta ei löytynyt, luodaan uusi.
    if kohde_b is None:
      kohde_b = LomakeB.Meta.model()
      # Asetetaan linkki B-->A, jos mahdollista
      try:
        setattr(kohde_b, avain_b, self.instance)
      except (AttributeError, TypeError):
        pass
    assert isinstance(kohde_b, LomakeB.Meta.model), (
      f'Kohde B ei voi olla tyyppiä {type(kohde_b)} != {LomakeB.Meta.model}!'
    )
    return {
      **({
        'data': self.data,
        'files': self.files,
      } if self.is_bound else {}),
      'initial': {
        avain.replace(tunnus + '-', '', 1): arvo
        for avain, arvo in self.initial.items()
        if avain.startswith(tunnus + '-') and avain != tunnus + '-'
      },
      'instance': kohde_b,
    }
    # def liitos_kwargs

  YhdistettyLomake = yhdista_lomakkeet.__wrapped__(
    LomakeA, LomakeB,
    tunnus=tunnus,
    liitos_kwargs=liitos_kwargs,
    pakollinen=pakollinen_b,
    valita_parametrit=valita_parametrit,
  )

  @functools.wraps(YhdistettyLomake, updated=())
  class YhdistettyLomake(YhdistettyLomake):
    # pylint: disable=function-redefined

    def __init__(self, *args, **kwargs):
      ''' Piilota ja estä mahdollinen paluuvierasavain B --> A. '''
      super().__init__(*args, **kwargs)
      lomake_b = getattr(self, tunnus)
      if avain_b in lomake_b.fields:
        lomake_b.fields[avain_b].disabled = True
        lomake_b.fields[avain_b].required = False
        lomake_b.fields[avain_b].widget = forms.HiddenInput()
      # def __init__

    # ModelForm:

    @transaction.atomic
    def _save_m2m(self):
      super()._save_m2m()
      # Jos viittaus A-->B voi olla tyhjä, tallennetaan B (vasta) nyt
      if not avain_a or self.Meta.model._meta.get_field(avain_a).null:

        # Haetaan A-kohde ja B-lomake
        kohde_a = self.instance
        lomake_b = getattr(self, tunnus)

        # Otetaan vanha B-kohde talteen
        vanha_kohde_b = getattr(kohde_a, avain_a) \
        if avain_a and hasattr(kohde_a, avain_a) else None

        if vanha_kohde_b == lomake_b.instance and vanha_kohde_b.pk:
          # Päivitetään olemassaoleva B
          try:
            with self.tallenna_liitos(tunnus):
              lomake_b.save(commit=True)
          except (ValueError, DatabaseError):
            if pakollinen_b:
              raise
            # Säilytetään olemassaoleva tietue.
            vanha_kohde_b.refresh_from_db()
          # if vanha_kohde_b == lomake_b.instance and vanha_kohde_b.pk
        else:
          # Asetetaan linkki B-->A
          setattr(lomake_b.instance, avain_b, kohde_a)

          # Yritetään tallentaa B ja otetaan virhe kiinni
          try:
            with self.tallenna_liitos(tunnus):
              kohde_b = lomake_b.save(commit=True)
          except (ValueError, DatabaseError):
            if pakollinen_b:
              raise
            kohde_b = vanha_kohde_b

          # Asetetaan tarvittaessa linkki A-->B ja tallennetaan A (uudelleen).
          if avain_a:
            setattr(kohde_a, avain_a, kohde_b)
            kohde_a.save()

          # Poistetaan vanha kohde, jos viittaus katkesi
          if vanha_kohde_b and vanha_kohde_b.pk and vanha_kohde_b != kohde_b:
            vanha_kohde_b.delete()
          # if vanha_kohde_b != lomake_b.instance:
        # if ...null

      # def _save_m2m

    @transaction.atomic
    def save(self, commit=True):
      # pylint: disable=access-member-before-definition
      # pylint: disable=attribute-defined-outside-init

      # Jos viittaus A-->B ei voi olla tyhjä, tallennetaan B ensin
      # (muussa tapauksessa B tallennetaan lopuksi `_save_m2m`-metodissa)
      if avain_a and not self.Meta.model._meta.get_field(avain_a).null:
        lomake_b = getattr(self, tunnus)
        with self.tallenna_liitos(tunnus):
          kohde_b = lomake_b.save(commit=commit)
        setattr(self.instance, avain_a, kohde_b)

        # Kun `commit=False`, ja B on uusi, tallentamaton tietokantarivi:
        if not commit and kohde_b and not kohde_b.pk:
          # Vaihdetaan A:n tallennusmetodi kertaluontoisesti.
          vanha_save = self.instance.save
          def save(instance):
            # Hae tämänhetkinen B, tallenna se ja aseta uudelleen.
            kohde_b = getattr(instance, avain_a)
            with self.tallenna_liitos(tunnus):
              kohde_b.save()
            # Päivitä B-lomakkeen rivi ja kutsu sen `save_m2m`-metodia.
            lomake_b.instance = kohde_b
            lomake_b.save_m2m()
            # Aseta B uudelleen A-kohteelle.
            setattr(instance, avain_a, kohde_b)
            # Palauta oletus-`save`-toteutus paikalleen ja kutsu sitä.
            instance.save = vanha_save
            instance.save()
            # def save
          self.instance.save = types.MethodType(save, self.instance)
          # if not commit and kohde_b and not kohde_b.pk
        return super().save(commit=commit)

      else:
        # Mikäli B on uusi, viittaus A.B on olemassa ja tämä voi olla
        # tyhjä, korvataan A.B arvolla None ennen A:n tallennusta.
        # Palautetaan tallennuksen jälkeen A.B = B.
        # B tallennetaan lopuksi `_save_m2m`-metodissa.
        # Django vaatii tämän A:n tallentamiseksi; ks.
        # django.db.models.base:Model._prepare_related_fields_for_save.
        lomake_b = getattr(self, tunnus)
        if avain_a is None or lomake_b.instance.pk is not None:
          return super().save(commit=commit)
        setattr(self.instance, avain_a, None)
        self.instance = super().save(commit=commit)
        setattr(self.instance, avain_a, lomake_b.instance)
        return self.instance
        # else
      # def save

    # class YhdistettyLomake

  return YhdistettyLomake
  # def yhdista_lomakkeet



# -*- coding: utf-8 -*-
# pylint: disable=invalid-name

from collections.abc import Iterable
import functools
import re

from django.db.models import ProtectedError
from django import forms
from django.template import loader
from django.utils.translation import gettext_lazy as _
from django import VERSION as django_versio

from .nidottu import yhdistetty_lomake


# Mallin nimen poiminta Djangon ProtectedError-poikkeusoliosta.
# Ks. django/db/models/deletion.py; `Collector.collect`-metodi.
mallin_nimi = re.compile(r'Cannot delete some instances of model ([^ ]+)')


@functools.wraps(yhdistetty_lomake)
def lisaa_lomakesarja(
  LomakeA, LomakeB,
  *,
  avain_a=None,
  avain_b=None,
  tunnus=None,
  epasuora=False,
  pakollinen=False,
  lomakesarja_parametrit=None,
  valita_parametrit: Iterable = (),
  **kwargs
):
  '''
  Yhdistää ModelForm-luokan ja toisesta ModelForm-luokasta
  muodostetun InlineFormSet-luokan silloin,
  kun B-luokasta on suora (ForeignKey) tai epäsuora (GenericForeignKey)
  viittaus A-luokkaan.

  Args:
    LomakeA, LomakeB: mallikohtaiset lomakeluokat
    avain_a, avain_b: viittaukset mallien välillä
    tunnus: yksikäsitteinen tunnus lomakesarjalle (oletus 'lomakesarja')
    epasuora: käytetään `generic_inlineformset_factory`-funktiota?
    lomakesarja_parametrit: parametrit, jotka asetetaan lomakesarjan määreiksi
    valita_parametrit: ulomman lomakkeen parametrit, jotka välitetään
      sellaisenaan lomakesarjalle
    *args, **kwargs: lisäparametrit ``inlineformset_factory``-funktiolle

  Returns:
    lomakeluokka
  '''
  # Pakotetaan ylimääräisten lomakkeiden määräksi nolla.
  kwargs['extra'] = 0

  assert tunnus or avain_a or avain_b, (
    'Joko `tunnus`, `avain_a` tai `avain_b` on määriteltävä.'
  )

  if epasuora:
    from django.contrib.contenttypes.forms import generic_inlineformset_factory
    lomakesarja = generic_inlineformset_factory(
      LomakeB.Meta.model,
      form=LomakeB,
      **kwargs
    )
  else:
    lomakesarja = forms.models.inlineformset_factory(
      LomakeA.Meta.model,
      LomakeB.Meta.model,
      form=LomakeB,
      **kwargs
    )

  # Käytetään tyhjän lomakkeen oletusarvoina
  # `initial`-datan ensimmäistä alkiota.
  class lomakesarja(lomakesarja):
    # pylint: disable=function-redefined
    def get_form_kwargs(self, index):
      if index is None and self.initial_extra:
        return {
          **super().get_form_kwargs(index), 'initial': self.initial_extra[0]
        }
      else:
        return super().get_form_kwargs(index)
      # def get_form_kwargs
    # class lomakesarja

  # Aseta mahdolliset lomakesarjan parametrit.
  for avain, arvo in (lomakesarja_parametrit or {}).items():
    setattr(lomakesarja, avain, arvo)

  # Lisää tarvittaessa oletusarvot HTML-piirtoa varten.
  if not hasattr(lomakesarja, 'label'):
    lomakesarja.label = (
      LomakeB.Meta.model._meta.verbose_name_plural
    ).capitalize()
  if not hasattr(lomakesarja, 'palikka'):
    lomakesarja.palikka = 'pumaska/lomakesarja_lomakekenttana.html'
  if not hasattr(lomakesarja, 'riviluokka'):
    lomakesarja.riviluokka = 'panel panel-default clearfix'
  if not hasattr(lomakesarja, 'lisaa_painike'):
    lomakesarja.lisaa_painike = _('Lisää %(malli)s') % {
      'malli': LomakeB.Meta.model._meta.verbose_name
    }
  if not hasattr(lomakesarja, 'poista_painike'):
    lomakesarja.poista_painike = _('Poista %(malli)s') % {
      'malli': LomakeB.Meta.model._meta.verbose_name
    }

  tunnus = tunnus or avain_a or (
    LomakeB.Meta.model._meta.get_field(avain_b).remote_field.name
  )

  if tunnus in LomakeA.base_fields:
    raise RuntimeError(
      f'Lomakesarjaa ei voida nitoa samalla nimellä'
      f' kuin olemassaoleva lomakekenttä: {tunnus}'
    )

  def liitos_kwargs(self: LomakeA):
    '''
    Korvataan oletusarvoinen, ulommasta initial-datasta poimittu
    `initial` luetteloksi käärityllä versiolla.

    Asetetaan ulompi olio lomakesarjan `instanceksi`.
    '''
    initial = {
      avain.replace(tunnus + '-', '', 1): arvo
      for avain, arvo in self.initial.items()
      if avain.startswith(tunnus + '-') and avain != tunnus + '-'
    }
    assert isinstance(self.instance, self.Meta.model), (
      f'Kohde A ei voi olla tyyppiä {type(self.instance)} != {self.Meta.model}!'
    )
    return {
      'initial': [initial] if initial else [],
      'instance': self.instance,
      'queryset': (
        getattr(self.instance, avain_a).all()
        if avain_a and self.instance.pk
        else None
      ),
    }
    # def liitos_kwargs

  YhdistettyLomake = lisaa_lomakesarja.__wrapped__(
    LomakeA, lomakesarja,
    tunnus=tunnus,
    liitos_kwargs=liitos_kwargs,
    valita_parametrit=valita_parametrit,
    pakollinen=pakollinen,
  )
  @functools.wraps(YhdistettyLomake, updated=())
  class YhdistettyLomake(YhdistettyLomake):
    # pylint: disable=function-redefined

    # HTML (Django 4-)

    if django_versio < (5, ):
      def _html_output(self, *args, **kwargs):
        '''
        Tuotetaan super-tuloste (A-lomake) sekä lomakesarjan perustoteutus:
        - kukin olemassaoleva B-lomake,
        - tyhjä B-lomake uuden lisäämiseksi,
        - hallintolomake ja
        - tarvittava Javascript rivien poistoon ja lisäämiseen.
        '''
        # pylint: disable=protected-access
        return super()._html_output(*args, **kwargs) \
        + loader.get_template('pumaska/lomakesarja.html').render({
          'tunnus': tunnus,
          'lomakesarja': getattr(self, tunnus),
        })
        # def _html_output


    # ModelForm

    def _save_m2m(self):
      '''
      Tallennetaan M2M-kohteet `super`-toteutuksen mukaisesti.
      Tämän jälkeen tallennetaan lomakesarja.

      Mikäli olemassaolevien rivien poisto epäonnistuu mallin tasolla
      `ProtectedError`-poikkeuksen takia, parsitaan tämän poikkeuksen
      viestitekstistä riippuvan mallin nimi ja tuotetaan vastaava
      `ValidationError`.
      '''
      super()._save_m2m()
      lomakesarja = getattr(self, tunnus)
      lomakesarja.instance = self.instance
      try:
        with self.tallenna_liitos(tunnus):
          lomakesarja.save(commit=True)
      except ProtectedError as exc:
        try:
          _mallin_nimi = mallin_nimi.match(str(exc))[1]
        except (TypeError, KeyError):
          exc2 = forms.ValidationError(str(exc), code='protect')
        else:
          exc2 = forms.ValidationError(
            _(
              'Rivin poisto epäonnistui:'
              ' suojattuja, riippuvia %(malli)s-kohteita.'
            ) % {'malli': _mallin_nimi},
            code='protect',
          )
        lomakesarja._non_form_errors.append(exc2)
        raise exc2 from exc
      # def _save_m2m

    # class YhdistettyLomake

  return YhdistettyLomake
  # def lisaa_lomakesarja



from collections.abc import Iterable
from contextlib import contextmanager
import functools
from itertools import chain

from django import forms
from django.utils.functional import cached_property
from django import VERSION as django_versio

from pumaska.piirto import Piirto


def yhdistetty_lomake(
  lomake: type,
  liitos: type = None,
  *,
  tunnus: str,
  liitos_kwargs: callable = None,
  pakollinen: bool = False,
  valita_parametrit: Iterable = (),
):
  assert valita_parametrit is not None
  if liitos is None:
    return functools.partial(
      yhdistetty_lomake,
      liitos=lomake,
      tunnus=tunnus,
      liitos_kwargs=liitos_kwargs,
      pakollinen=pakollinen,
      valita_parametrit=valita_parametrit
    )
    # if liitos is None

  class YhdistettyLomake(lomake):

    def __init__(self, *args, prefix=None, **kwargs):
      ajonaikainen_kwargs = kwargs.pop(f'{tunnus}_kwargs', {})
      asetettavat_maareet = {}
      for param in valita_parametrit:
        try:
          ajonaikainen_kwargs[param] = kwargs[param]
        except KeyError:
          pass

      super().__init__(*args, prefix=prefix, **kwargs)

      # Käytetään A-lomakkeen vedostajaa oletuksena myös B-lomakkeella.
      # Huomaa:
      # - Django vaatii, että `renderer` periytyy lomakeluokan
      #   mahdollisesta `default_renderer`-luokasta;
      # - lomake ei ota (Django 4.2) vastaan parametriä `renderer`,
      #   tämä täytyy asettaa jälkeenpäin määreenä.
      if not 'renderer' in ajonaikainen_kwargs:
        class PiirtoB(Piirto, LomakeB=liitos): pass
        # Lomakesarja ei ota `renderer`-parametriä vastaan.
        # Asetetaan tämä alustuksen jälkeen määreenä.
        # Huomaa, että Django 5.0 käyttää erillistä `form_renderer`-määrettä
        # lomakesarjaan sisältyvien lomakkeiden piirtoon.
        if issubclass(liitos, forms.BaseFormSet):
          asetettavat_maareet['renderer'] \
          = asetettavat_maareet['form_renderer'] = PiirtoB(self)
        else:
          ajonaikainen_kwargs['renderer'] = PiirtoB(self)
        # if not 'renderer' in ajonaikainen_kwargs

      # Annetaan sisemmän lomakkeen alustuksessa oletuksena:
      # - data ja files, mikäli ulompi lomake on lähetetty,
      # - initial: `<tunnus>-`-alkuiset, epätyhjät avaimet ulomman
      #   lomakkeen initial-datassa
      # - prefix: `<ulompi prefix>-<tunnus>`
      # Huomaa, että `liitos_kwargs()` ja `ajonaikainen_kwargs` (tässä
      # järjestyksessä) ylikirjoittavat nämä oletusarvot.
      _liitos = liitos(**{
        **({
          'data': self.data,
          'files': self.files,
        } if self.is_bound else {}),
        'initial': {
          avain.replace(tunnus + '-', '', 1): arvo
          for avain, arvo in self.initial.items()
          if avain.startswith(tunnus + '-') and avain != tunnus + '-'
        },
        'prefix': f'{self.prefix}-{tunnus}' if self.prefix else tunnus,
        **(
          liitos_kwargs(self)
          if callable(liitos_kwargs)
          else liitos_kwargs if liitos_kwargs is not None
          else {}
        ),
        **ajonaikainen_kwargs,
      })

      # Asetetaan tarvittavat määreet (renderer).
      for avain, arvo in asetettavat_maareet.items():
        setattr(_liitos, avain, arvo)

      # Asetetaan viittaus ulommasta lomakkeesta sisempään.
      setattr(self, tunnus, _liitos)

      # Jos B-viittaus saa olla tyhjä:
      # - asetetaan kaikki B-lomakkeen kentät valinnaisiksi GET-pyynnöllä;
      # – huomaa, että tämä koskee myös mahdollisten sisäkkäisten
      #   lomakkeiden (C) kenttiä; ks. `__iter__`-toteutus alla;
      # - ohitetaan vimpainten `required`-määreen tulostus;
      # - huomaa, että tämä ei tee mitään sisemmälle lomakesarjalle.
      if not pakollinen:
        for kentta in _liitos:
          if hasattr(kentta, 'field'):
            if not self.data:
              kentta.field.required = False
            kentta.field.widget.use_required_attribute = lambda initial: False
      # def __init__

    # def order_fields(self, field_order)
    # def __str__(self)
    # def __repr__(self)

    def __iter__(self):
      '''
      Iteroidaan lomakesarjaan liittyvät lomakkeet erikseen; lomake erikseen.
      '''
      _liitos = getattr(self, tunnus)
      return chain(
        super().__iter__(),
        chain.from_iterable(_liitos)
        if isinstance(_liitos, forms.BaseFormSet)
        else _liitos.__iter__(),
      )
      # def __iter__

    def __getitem__(self, item):
      '''
      Poimi liitoslomake(sarja), liitoslomakkeen kenttä tai (lomakesarjan
      tapauksessa) liitoslomakesarjan lomake tai tämän kenttä.
      '''
      if item == tunnus:
        return getattr(self, tunnus)
      elif item.startswith(f'{tunnus}-') \
      and (_liitos_kentta := item.partition(f'{tunnus}-')[2]):
        _liitos = getattr(self, tunnus)
        if isinstance(_liitos, forms.BaseFormSet):
          indeksi, _, item = _liitos_kentta.partition('-')
          _liitos_lomake = _liitos.__getitem__(int(indeksi))
          if item:
            return _liitos_lomake.__getitem__(item)
          else:
            return _liitos_lomake
        else:
          return _liitos.__getitem__(_liitos_kentta)
      else:
        return super().__getitem__(item)
      # def __getitem__

    @property
    def errors(self):
      '''
      Lisää liitoslomakkeen tai -sarjan mahdolliset virheet silloin, kun
      liitos on pakollinen tai sitä on muokattu.
      '''
      virheet = list(super().errors.items())
      if (_liitos := getattr(self, tunnus)).has_changed() \
      or pakollinen:
        if isinstance(_liitos, forms.BaseFormSet):
          for indeksi, lomake in enumerate(_liitos.forms):
            if lomake not in _liitos.deleted_forms:
              for avain, arvo in list(lomake.errors.items()):
                virheet.append([
                  '%s-%d-%s' % (tunnus, indeksi, avain), arvo
                ])
          if any(_liitos.non_form_errors()):
            virheet.append([
              # Lisää lomakeriippumattomat virheet hallintolomakkeen kohdalle.
              tunnus + '-TOTAL_FORMS',
              _liitos.non_form_errors()
            ])
        else:
          for avain, arvo in list(_liitos.errors.items()):
            virheet.append([
              '%s-%s' % (tunnus, avain), arvo
            ])
      return forms.utils.ErrorDict(virheet)
      # def errors

    def is_valid(self):
      '''
      Jos B-viittaus saa olla tyhjä eikä sitä ole muokattu,
      ei välitetä B-lomakkeen mahdollisesta epäkelpoisuudesta.
      '''
      return super().is_valid() and (
        (_liitos := getattr(self, tunnus)).is_valid()
        or (not pakollinen and not _liitos.has_changed())
      )
      # def is_valid

    # def add_prefix(self, field_name)
    # def add_initial_prefix(self, field_name)

    if django_versio < (5, ):
      def _html_output(self, *args, **kwargs):
        # pylint: disable=protected-access
        return super()._html_output(*args, **kwargs) \
        + getattr(self, tunnus)._html_output(*args, **kwargs)
        # def _html_output

    # def as_table(self)
    # def as_ul(self)
    # def as_p(self)
    # def non_field_errors(self)
    # def add_error(self, field, error)
    # def has_error(self, field, code=None)
    # def full_clean(self)
    # def _clean_fields(self)
    # def _clean_form(self)
    # def _post_clean(self)
    # def clean(self)

    def has_changed(self):
      return super().has_changed() \
      or getattr(self, tunnus).has_changed()
      # def has_changed

    @cached_property
    def changed_data(self):
      '''
      Palauta ylälomakkeen omien muutosten lisäksi
      liitoslomakkeen mahdolliset muutokset
      `tunnus`-etuliitteellä varustettuina.
      '''
      muutokset = super().changed_data
      _liitos = getattr(self, tunnus)
      if isinstance(_liitos, forms.BaseFormSet):
        muutokset += [
          # Muodosta lomakekohtainen kentän etuliite poistamalla
          # liitetyn lomakkeen `prefixin` alusta
          # käsillä olevan (ylä-) lomakkeen oma `prefix` ja välimerkki -.
          f'{lomakekohtainen_tunnus}-{kentta}'
          for lomake, lomakekohtainen_tunnus in (
            (
              lomake,
              (
                lomake.prefix.replace(self.prefix + "-", "", 1)
              ) if self.prefix else lomake.prefix
            )
            for lomake in _liitos
          )
          for kentta in lomake.changed_data
        ]
      else:
        muutokset += [
          f'{tunnus}-{kentta}'
          for kentta in getattr(_liitos, 'changed_data', ())
        ]
      return muutokset
      # def changed_data

    @property
    def media(self):
      return super().media + getattr(self, tunnus).media

    #def is_multipart(self)

    def hidden_fields(self):
      return [
        f for f in super().hidden_fields()
        if f.form is self
      ]
      # def hidden_fields

    #def visible_fields(self)
    #def get_initial_for_field(self, field, field_name)

    # `in`

    def __contains__(self, key):
      if key == tunnus:
        return True
      elif key.startswith(f'{tunnus}-'):
        key = key.partition(f'{tunnus}-')[2]
        if hasattr(getattr(self, tunnus), '__contains__') \
        and getattr(self, tunnus).__contains__(key):
          return True
        return key in getattr(self, tunnus).fields
        # if key.startswith
      elif hasattr(super(), '__contains__') \
      and super().__contains__(key):
        return True
      else:
        return key in self.fields
      # def __contains__

    @staticmethod
    @contextmanager
    def tallenna_liitos(tunnus):
      ''' Lisätään `tunnus` etuliitteenä mahdollisiin virheisiin. '''
      try:
        yield
      except forms.ValidationError as exc:
        try:
          sanakirja = exc.message_dict
        except AttributeError:
          raise type(exc)({tunnus: exc}) from exc
        else:
          raise type(exc)({
            '-'.join((tunnus, avain)): arvo
            for avain, arvo in sanakirja.items()
          }) from exc
      # def tallenna_liitos

    # class YhdistettyLomake

  YhdistettyLomake.__name__ += f'_{tunnus}'
  YhdistettyLomake.__qualname__ += f'_{tunnus}'
  return YhdistettyLomake
  # def yhdista_lomakkeet



# Generated by Django 5.1.dev20231216113328 on 2023-12-18 11:12

import django.db.models.deletion
from django.db import migrations, models


class Migration(migrations.Migration):

    initial = True

    dependencies = [
    ]

    operations = [
        migrations.CreateModel(
            name='Asiakas',
            fields=[
                ('id', models.BigAutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('nimi', models.CharField(max_length=255)),
            ],
        ),
        migrations.CreateModel(
            name='Edustaja',
            fields=[
                ('id', models.BigAutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('nimi', models.CharField(max_length=255)),
            ],
        ),
        migrations.CreateModel(
            name='Paamies',
            fields=[
                ('id', models.BigAutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('nimi', models.CharField(max_length=255)),
            ],
        ),
        migrations.CreateModel(
            name='Asiakasosoite',
            fields=[
                ('id', models.BigAutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('osoite', models.CharField(max_length=255)),
                ('asiakas', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, to='testit.asiakas')),
            ],
        ),
        migrations.AddField(
            model_name='asiakas',
            name='edustaja',
            field=models.OneToOneField(blank=True, null=True, on_delete=django.db.models.deletion.SET_NULL, to='testit.edustaja'),
        ),
        migrations.CreateModel(
            name='Lasku',
            fields=[
                ('id', models.BigAutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('numero', models.IntegerField()),
                ('asiakas', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, to='testit.asiakas')),
                ('paamies', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, to='testit.paamies')),
            ],
        ),
        migrations.CreateModel(
            name='Rivi',
            fields=[
                ('id', models.BigAutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('summa', models.DecimalField(decimal_places=2, max_digits=11)),
                ('lasku', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, to='testit.lasku')),
            ],
        ),
    ]



class PiirtoMeta(type):
  def __new__(mcs, name, bases, attrs, *, LomakeB):
    if getattr(LomakeB, 'default_renderer', False):
      bases = (LomakeB.default_renderer, *bases)
    return super().__new__(mcs, name, bases, attrs)
  # class PiirtoMeta


class Piirto(metaclass=PiirtoMeta, LomakeB=None):
  ''' Hae kaikki jäsenet alkuperäisen lomakkeen `rendereriltä`. '''

  __slots__ = ['_lomake_a']

  def __init__(self, lomake_a):
    self._lomake_a = lomake_a

  def __getattribute__(self, attr):
    try:
      return super().__getattribute__(attr)
    except:
      return self._lomake_a.renderer.__getattribute__(attr)
    # def __getattribute__

  # class Piirto
